/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "port.h"
#include "cheats.h"

#include "mac-prefix.h"
#include "mac-dialog.h"
#include "mac-os.h"
#include "mac-stringtools.h"
#include "mac-cheat.h"

#define	kDataBrowser	'BRSR'
#define	kCmCheckBox		'CHK_'
#define	kCmAddress		'ADDR'
#define	kCmValue		'VALU'
#define	kCmDescription	'DESC'
#define	kNewButton		'NEW_'
#define	kDelButton		'DEL_'
#define	kAllButton		'ALL_'

extern SCheatData	Cheat;

static void InitCheatItems(void);
static void ImportCheatItems(void);
static void DetachCheatItems(void);
static void AddCheatItem(void);
static void DeleteCheatItem(void);
static void EnableAllCheatItems(void);
static pascal void DBItemNotificationCallBack(ControlRef, DataBrowserItemID, DataBrowserItemNotification);
static pascal Boolean DBCompareCallBack(ControlRef, DataBrowserItemID, DataBrowserItemID, DataBrowserPropertyID);
static pascal OSStatus DBClientDataCallback(ControlRef, DataBrowserItemID, DataBrowserPropertyID, DataBrowserItemDataRef, Boolean);
static pascal OSStatus CheatEventHandler(EventHandlerCallRef, EventRef, void *);

typedef struct
{
	UInt32	id;
	Boolean	valid;
	Boolean	enabled;
	UInt32	address;
	UInt8	value;
	char	description[22];
}	CheatItem;

static ControlRef	dbRef;
static WindowRef	wRef;
static CheatItem	citem[MAX_CHEATS];
static UInt32		numofcheats;

static void InitCheatItems(void)
{
	for (UInt32 i = 0; i < MAX_CHEATS; i++)
	{
		citem[i].id      = i + 1;
		citem[i].valid   = false;
		citem[i].enabled = false;
		citem[i].address = 0;
		citem[i].value   = 0;
		sprintf(citem[i].description, "Cheat %02lu", citem[i].id);
	}
}

static void ImportCheatItems(void)
{
	for (UInt32 i = 0; i < Cheat.num_cheats; i++)
	{
		citem[i].valid   = true;
		citem[i].enabled = Cheat.c[i].enabled;
		citem[i].address = Cheat.c[i].address;
		citem[i].value   = Cheat.c[i].byte;
		strcpy(citem[i].description, Cheat.c[i].name);
	}
}

static void DetachCheatItems(void)
{
	S9xDeleteCheats(); // ->  include Cheat.num_cheats = 0
	
	for (UInt32 i = 0; i < MAX_CHEATS; i++)
	{
		if (citem[i].valid)
		{
			strcpy(Cheat.c[Cheat.num_cheats].name, citem[i].description);
			S9xAddCheat(citem[i].enabled, false, citem[i].address, citem[i].value); // ->  include Cheat.num_cheats++
		}
	}
	
	S9xApplyCheats();
}

void ConfigureCheat(void)
{
	OSStatus						err;
	IBNibRef						nibRef;
	CGrafPtr						old;
	ControlRef						control;
	ControlID						cID;
	DataBrowserCallbacks			callbacks;
	DataBrowserListViewColumnDesc	columnDesc;
	EventHandlerRef					eref;
	EventHandlerUPP					eventUPP;
	EventTypeSpec					windowEvents[] = { { kEventClassCommand, kEventCommandProcess }, { kEventClassWindow, kEventWindowClose } };
	Rect							rct = { 20, 20, 252, 416 };

	if (!cartOpen)
		return;
	
	err = CreateNibReference(CFSTR("SNES9X"), &nibRef);
	if (err)
		return;
	
	err = CreateWindowFromNib(nibRef, CFSTR("CheatEntry"), &wRef);
	if (err)
		return;
	
	GetPort(&old);
	SetPortWindowPort(wRef);
	
	err = CreateDataBrowserControl(wRef, &rct, kDataBrowserListView, &dbRef);

	columnDesc.propertyDesc.propertyID = kCmCheckBox;
	columnDesc.propertyDesc.propertyType = kDataBrowserCheckboxType;
	columnDesc.propertyDesc.propertyFlags = kDataBrowserListViewSortableColumn | kDataBrowserListViewSelectionColumn | kDataBrowserPropertyIsMutable | kDataBrowserDefaultPropertyFlags;
	columnDesc.headerBtnDesc.version = kDataBrowserListViewLatestHeaderDesc;
	columnDesc.headerBtnDesc.minimumWidth = 30;
	columnDesc.headerBtnDesc.maximumWidth = 30;
	columnDesc.headerBtnDesc.titleOffset = 0;
	columnDesc.headerBtnDesc.titleString = nil;
	columnDesc.headerBtnDesc.initialOrder = kDataBrowserOrderIncreasing;
	columnDesc.headerBtnDesc.btnFontStyle.flags = kControlUseJustMask;
	columnDesc.headerBtnDesc.btnFontStyle.just = teCenter;
	columnDesc.headerBtnDesc.btnContentInfo.contentType = 0;
	columnDesc.headerBtnDesc.btnContentInfo.u.resID = 0;

	err = AddDataBrowserListViewColumn(dbRef, &columnDesc, ULONG_MAX);
	
	columnDesc.propertyDesc.propertyID = kCmAddress;
	columnDesc.propertyDesc.propertyType = kDataBrowserTextType;
	columnDesc.propertyDesc.propertyFlags = kDataBrowserListViewSortableColumn | kDataBrowserListViewSelectionColumn | kDataBrowserPropertyIsMutable | kDataBrowserDefaultPropertyFlags;
	columnDesc.headerBtnDesc.version = kDataBrowserListViewLatestHeaderDesc;
	columnDesc.headerBtnDesc.minimumWidth = 84;
	columnDesc.headerBtnDesc.maximumWidth = 84;
	columnDesc.headerBtnDesc.titleOffset = 0;
	columnDesc.headerBtnDesc.titleString = CFCopyLocalizedString(CFSTR("Address"), "address");
	columnDesc.headerBtnDesc.initialOrder = kDataBrowserOrderIncreasing;
	columnDesc.headerBtnDesc.btnFontStyle.flags = kControlUseJustMask;
	columnDesc.headerBtnDesc.btnFontStyle.just = teCenter;
	columnDesc.headerBtnDesc.btnContentInfo.contentType = 0;
	columnDesc.headerBtnDesc.btnContentInfo.u.resID = 0;
	
	err = AddDataBrowserListViewColumn(dbRef, &columnDesc, ULONG_MAX);
	CFRelease(columnDesc.headerBtnDesc.titleString);
	
	columnDesc.propertyDesc.propertyID = kCmValue;
	columnDesc.propertyDesc.propertyType = kDataBrowserTextType;
	columnDesc.propertyDesc.propertyFlags = kDataBrowserListViewSortableColumn | kDataBrowserListViewSelectionColumn | kDataBrowserPropertyIsMutable | kDataBrowserDefaultPropertyFlags;
	columnDesc.headerBtnDesc.version = kDataBrowserListViewLatestHeaderDesc;
	columnDesc.headerBtnDesc.minimumWidth = 65;
	columnDesc.headerBtnDesc.maximumWidth = 65;
	columnDesc.headerBtnDesc.titleOffset = 0;
	columnDesc.headerBtnDesc.titleString = CFCopyLocalizedString(CFSTR("Value"), "value");
	columnDesc.headerBtnDesc.initialOrder = kDataBrowserOrderIncreasing;
	columnDesc.headerBtnDesc.btnFontStyle.flags = kControlUseJustMask;
	columnDesc.headerBtnDesc.btnFontStyle.just = teCenter;
	columnDesc.headerBtnDesc.btnContentInfo.contentType = 0;
	columnDesc.headerBtnDesc.btnContentInfo.u.resID = 0;
	
	err = AddDataBrowserListViewColumn(dbRef, &columnDesc, ULONG_MAX);
	CFRelease(columnDesc.headerBtnDesc.titleString);

	columnDesc.propertyDesc.propertyID = kCmDescription;
	columnDesc.propertyDesc.propertyType = kDataBrowserTextType;
	columnDesc.propertyDesc.propertyFlags = kDataBrowserListViewSortableColumn | kDataBrowserListViewSelectionColumn | kDataBrowserPropertyIsMutable | kDataBrowserDefaultPropertyFlags;
	columnDesc.headerBtnDesc.version = kDataBrowserListViewLatestHeaderDesc;
	columnDesc.headerBtnDesc.minimumWidth = 196;
	columnDesc.headerBtnDesc.maximumWidth = 196;
	columnDesc.headerBtnDesc.titleOffset = 0;
	columnDesc.headerBtnDesc.titleString = CFCopyLocalizedString(CFSTR("Description"), "description");
	columnDesc.headerBtnDesc.initialOrder = kDataBrowserOrderIncreasing;
	columnDesc.headerBtnDesc.btnFontStyle.flags = kControlUseJustMask;
	columnDesc.headerBtnDesc.btnFontStyle.just = teFlushLeft;
	columnDesc.headerBtnDesc.btnContentInfo.contentType = 0;
	columnDesc.headerBtnDesc.btnContentInfo.u.resID = 0;
	
	err = AddDataBrowserListViewColumn(dbRef, &columnDesc, ULONG_MAX);
	CFRelease(columnDesc.headerBtnDesc.titleString);

	callbacks.version = kDataBrowserLatestCallbacks;
	err = InitDataBrowserCallbacks(&callbacks);
	callbacks.u.v1.itemDataCallback = NewDataBrowserItemDataUPP(DBClientDataCallback);
	callbacks.u.v1.itemCompareCallback = NewDataBrowserItemCompareUPP(DBCompareCallBack);
	callbacks.u.v1.itemNotificationCallback = NewDataBrowserItemNotificationUPP(DBItemNotificationCallBack);
	err = SetDataBrowserCallbacks(dbRef, &callbacks);

    err = SetDataBrowserHasScrollBars(dbRef, false, true);
	err = SetDataBrowserSortOrder(dbRef, kDataBrowserOrderIncreasing);
	err = SetDataBrowserSortProperty(dbRef, kCmAddress);
	err = SetDataBrowserSelectionFlags(dbRef, kDataBrowserDragSelect | kDataBrowserCmdTogglesSelection);
	
	InitCheatItems();
	ImportCheatItems();
	
	DataBrowserItemID	*id;
	
	id = new DataBrowserItemID [MAX_CHEATS];
	if (!id)
		QuitWithFatalError(0, "cheat 01");
	
	numofcheats = 0;

	for (UInt32 i = 0; i < MAX_CHEATS; i++)
	{
		if (citem[i].valid)
		{
			id[numofcheats] = citem[i].id;
			numofcheats++;
		}
	}
	
	if (numofcheats)
		err = AddDataBrowserItems(dbRef, kDataBrowserNoItem, numofcheats, id, kDataBrowserItemNoProperty);
		
	delete [] id;
	
	cID.id = 0;
	
	cID.signature = kNewButton;
	GetControlByID(wRef, &cID, &control);
	if (numofcheats == MAX_CHEATS)
		err = DeactivateControl(control);
	else
		err = ActivateControl(control);
		
	cID.signature = kAllButton;
	GetControlByID(wRef, &cID, &control);
	if (numofcheats == 0)
		err = DeactivateControl(control);
	else
		err = ActivateControl(control);

	cID.signature = kDelButton;
	GetControlByID(wRef, &cID, &control);
	err = DeactivateControl(control);

	err = InstallStandardEventHandler(GetWindowEventTarget(wRef));
	eventUPP = NewEventHandlerUPP(CheatEventHandler);
	err = InstallWindowEventHandler(wRef, eventUPP, 2, windowEvents, (void *) wRef, &eref);

	AdvanceKeyboardFocus(wRef);

	MoveWindowPosition(wRef, kWindowCheatEntry);
	ShowWindow(wRef);
	err = RunAppModalLoopForWindow(wRef);
	HideWindow(wRef);
	SaveWindowPosition(wRef, kWindowCheatEntry);

	err = RemoveEventHandler(eref);
	DisposeEventHandlerUPP(eventUPP);
	
	DisposeDataBrowserItemNotificationUPP(callbacks.u.v1.itemNotificationCallback);
	DisposeDataBrowserItemCompareUPP(callbacks.u.v1.itemCompareCallback);
	DisposeDataBrowserItemDataUPP(callbacks.u.v1.itemDataCallback);
	
	DisposeControl(dbRef);
	DisposeWindow(wRef);
	DisposeNibReference(nibRef);

	SetPort(old);
	
	DetachCheatItems();
}

static void AddCheatItem(void)
{
	OSStatus			err;
	ControlRef			control;
	ControlID			cid;
	UInt32				i;
	DataBrowserItemID	id[1];

	if (numofcheats == MAX_CHEATS)
		return;
	
	for (i = 0; i < MAX_CHEATS; i++)
		if (citem[i].valid == false)
			break;
			
	if (i == MAX_CHEATS)
		return;
		
	numofcheats++;
	citem[i].valid   = true;
	citem[i].enabled = false;
	citem[i].address = 0;
	citem[i].value   = 0;
	sprintf(citem[i].description, "Cheat %02lu", citem[i].id);
	
	id[0] = citem[i].id;
	err = AddDataBrowserItems(dbRef, kDataBrowserNoItem, 1, id, kDataBrowserItemNoProperty);
	err = RevealDataBrowserItem(dbRef, id[0], kCmAddress, true);
	
	cid.id = 0;
	
	if (numofcheats == MAX_CHEATS)
	{
		cid.signature = kNewButton;		
		GetControlByID(wRef, &cid, &control);
		err = DeactivateControl(control);
	}

	if (numofcheats)
	{
		cid.signature = kAllButton;		
		GetControlByID(wRef, &cid, &control);
		err = ActivateControl(control);
	}
}

static void DeleteCheatItem(void)
{
	OSStatus	err;
	ControlRef	control;
	ControlID	cid;
	Handle		selectedItems;
	UInt32		selectionCount;
	
	selectedItems = NewHandle(0);
	if (!selectedItems)
		return;
		
	err = GetDataBrowserItems(dbRef, kDataBrowserNoItem, true, kDataBrowserItemIsSelected, selectedItems);
	selectionCount = (GetHandleSize(selectedItems) / sizeof(DataBrowserItemID));

	if (selectionCount == 0)
	{
		DisposeHandle(selectedItems);
		return;
	}
	
	err = RemoveDataBrowserItems(dbRef, kDataBrowserNoItem, selectionCount, (DataBrowserItemID *) *selectedItems, kDataBrowserItemNoProperty);
	
	for (UInt32 i = 0; i < selectionCount; i++)
	{
		citem[((DataBrowserItemID *) (*selectedItems))[i] - 1].valid   = false;
		citem[((DataBrowserItemID *) (*selectedItems))[i] - 1].enabled = false;
		numofcheats--;
	}
	
	DisposeHandle(selectedItems);
	
	cid.id = 0;
	
	if (numofcheats < MAX_CHEATS)
	{
		cid.signature = kNewButton;		
		GetControlByID(wRef, &cid, &control);
		err = ActivateControl(control);
	}
	
	if (numofcheats == 0)
	{
		cid.signature = kAllButton;		
		GetControlByID(wRef, &cid, &control);
		err = DeactivateControl(control);
	}
}

static void EnableAllCheatItems(void)
{
	OSStatus	err;
	
	for (int i = 0; i < MAX_CHEATS; i++)
		if (citem[i].valid)
			citem[i].enabled = true;
	
	err = UpdateDataBrowserItems(dbRef, kDataBrowserNoItem, kDataBrowserNoItem, nil, kDataBrowserItemNoProperty, kCmCheckBox);
}	

static pascal OSStatus DBClientDataCallback(ControlRef browser, DataBrowserItemID itemID, DataBrowserPropertyID property, DataBrowserItemDataRef itemData, Boolean changeValue)
{
	OSStatus 	err, result;
	CFStringRef	str;
	Boolean		r;
	uint32		address;
	uint8		value;		
	char		code[256];
	
	result = noErr;
	
	switch (property)
	{
		case kCmCheckBox:
			ThemeButtonValue	buttonValue;

            if (changeValue)
			{
				err = GetDataBrowserItemDataButtonValue(itemData, &buttonValue);
				citem[itemID - 1].enabled = (buttonValue == kThemeButtonOn) ? true : false;
 	        }
			else
				err = SetDataBrowserItemDataButtonValue(itemData, citem[itemID - 1].enabled ? kThemeButtonOn : kThemeButtonOff);
				
			break;

		case kCmAddress:
			if (changeValue)
			{
				err = GetDataBrowserItemDataText(itemData, &str);
				r = CFStringGetCString(str, code, 256, CFStringGetSystemEncoding());
				CFRelease(str);
				if (r)
				{	
					Boolean	translated;
			
					if (S9xProActionReplayToRaw(code, address, value) == nil)
						translated = true;
					else
					if (S9xGameGenieToRaw(code, address, value) == nil)
						translated = true;
					else
					{
						translated = false;
						if (sscanf(code, "%lx", &address) != 1)
							address = 0;
					}
						
					citem[itemID - 1].address = address;
					sprintf(code, "%06lX", address);
					str = CFStringCreateWithCString(kCFAllocatorDefault, code, CFStringGetSystemEncoding());
					err = SetDataBrowserItemDataText(itemData, str);
					CFRelease(str);
					
					if (translated)
					{
						DataBrowserItemID	id[1];
						
						citem[itemID - 1].value = value;
						id[0] = itemID;
						err = UpdateDataBrowserItems(browser, kDataBrowserNoItem, 1, id, kDataBrowserItemNoProperty, kCmValue);
					}
				}
			}	
			else
			{
				sprintf(code, "%06lX", citem[itemID - 1].address);
				str = CFStringCreateWithCString(kCFAllocatorDefault, code, CFStringGetSystemEncoding());
				err = SetDataBrowserItemDataText(itemData, str);
				CFRelease(str);
			}

			break;
			
		case kCmValue:
			if (changeValue)
			{
				err = GetDataBrowserItemDataText(itemData, &str);
				r = CFStringGetCString(str, code, 256, CFStringGetSystemEncoding());
				CFRelease(str);
				if (r)
				{
					UInt32	byte;
					
					if (sscanf(code, "%lx", &byte) == 1)
						citem[itemID - 1].value = (UInt8) byte;
					else
					{
						citem[itemID - 1].value = 0;
						err = SetDataBrowserItemDataText(itemData, CFSTR("00"));
					}
				}
			}	
			else
			{
				sprintf(code, "%02X", citem[itemID - 1].value);
				str = CFStringCreateWithCString(kCFAllocatorDefault, code, CFStringGetSystemEncoding());
				err = SetDataBrowserItemDataText(itemData, str);
				CFRelease(str);
			}

			break;
			
		case kCmDescription:
			if (changeValue)
			{
				err = GetDataBrowserItemDataText(itemData, &str);
				r = CFStringGetCString(str, code, 256, CFStringGetSystemEncoding());
				if (r)
				{
					CFRelease(str);

					if (strlen(code) > 19)
					{
						code[19] = 0;
						if (CharacterByteType(code, 18, smSystemScript) == smFirstByte)
							code[18] = 0;
					}
					else
					if (strlen(code) == 0)
					{
						code[0] = ' ';
						code[1] = 0;
					}

					strcpy(citem[itemID - 1].description, code);
				}
			}
			else
			{	
				str = CFStringCreateWithCString(kCFAllocatorDefault, citem[itemID - 1].description, CFStringGetSystemEncoding());
				err = SetDataBrowserItemDataText(itemData, str);
				CFRelease(str);
			}
			
			break;
		
		case kDataBrowserItemIsActiveProperty:
			err = SetDataBrowserItemDataBooleanValue(itemData, true);
			break;
		
		case kDataBrowserItemIsEditableProperty:
			err = SetDataBrowserItemDataBooleanValue(itemData, true);
			break;
		
		default:
			result = errDataBrowserPropertyNotSupported;
	}
	
	return result;
}

static pascal Boolean DBCompareCallBack(ControlRef browser, DataBrowserItemID itemOne, DataBrowserItemID itemTwo, DataBrowserPropertyID sortProperty)
{
	Boolean	result = false;
	
	switch (sortProperty)
	{
		case kCmCheckBox:
			result = (citem[itemOne - 1].enabled && !(citem[itemTwo - 1].enabled)) ? true : false;
			break;
			
		case kCmAddress:
			result = (citem[itemOne - 1].address <    citem[itemTwo - 1].address)  ? true : false;
			break;
			
		case kCmValue:
			result = (citem[itemOne - 1].value   <    citem[itemTwo - 1].value)    ? true : false;
			break;
		
		case kCmDescription:
			result = (strcmp(citem[itemOne - 1].description, citem[itemTwo - 1].description) < 0) ? true : false;
	}
	
	return result;
}

static pascal void DBItemNotificationCallBack(ControlRef browser, DataBrowserItemID itemID, DataBrowserItemNotification message)
{
	OSStatus	err;
	ControlRef	control;
	ControlID	cid = { kDelButton, 0 };
	UInt32		selectionCount;
	
	switch (message)
	{
		case kDataBrowserSelectionSetChanged:
			GetControlByID(wRef, &cid, &control);

			err = GetDataBrowserItemCount(browser, kDataBrowserNoItem, true, kDataBrowserItemIsSelected, &selectionCount);
			if (selectionCount == 0)
				err = DeactivateControl(control);
			else
				err = ActivateControl(control);
	}
}

static pascal OSStatus CheatEventHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData)
{
	#pragma unused (inHandlerCallRef)
	
	OSStatus	err, result = eventNotHandledErr;
	WindowRef	tWindowRef;

	tWindowRef = (WindowRef) inUserData;
	
	switch (GetEventClass(inEvent))
	{
		case kEventClassWindow:
			switch (GetEventKind(inEvent))
			{
				case kEventWindowClose:
					QuitAppModalLoopForWindow(tWindowRef);
					result = noErr;
			}
			
			break;

		case kEventClassCommand:
			switch (GetEventKind(inEvent))
			{
				case kEventCommandProcess:
					HICommand	tHICommand;

					err = GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, nil, sizeof(HICommand), nil, &tHICommand);
					if (err == noErr)
					{
						switch (tHICommand.commandID)
						{
							case kNewButton:
								AddCheatItem();
								result = noErr;
								break;
								
							case kDelButton:
								DeleteCheatItem();
								result = noErr;
								break;
								
							case kAllButton:
								EnableAllCheatItems();
								result = noErr;
						}
					}
			}
	}
	
	return result;
}
