/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "port.h"

#include "mac-prefix.h"
#include "mac-dialog.h"
#include "mac-gworld.h"
#include "mac-os.h"
#include "mac-keyboard.h"

#define kmUpArrowKey	0x7E
#define kmDownArrowKey	0x7D
#define kmRightArrowKey	0x7C
#define kmLeftArrowKey	0x7B
#define kmReturnKey		0x24
#define kmTabKey		0x30
#define kmShiftKey		0x38
#define kmControlKey	0x3B
#define kmOptionKey		0x3A
#define kmCommandKey	0x37
#define kmXKey			0x07
#define kmZKey			0x06
#define kmKP2Key		0x54
#define kmKP4Key		0x56
#define kmKP5Key		0x57
#define kmKP6Key		0x58
#define kmKP8Key		0x5B
#define kmKPEnterKey	0x4C
#define kmKPPlusKey		0x45
#define kmKP0Key		0x52
#define kmKPPeriodKey	0x41
#define kmHomeKey		0x73
#define kmPageUpKey		0x74
#define kmEndKey		0x77
#define kmPageDownKey	0x79
#define kmBackslashKey	0x2A 
#define km1Key			0x12
#define km0Key			0x1D
#define kmIKey			0x22
#define kmJKey			0x26
#define kmKKey			0x28
#define kmLKey			0x25
#define kmTildeKey		0x32
#define	kmRKey			0x0F
#define	kmBKey			0x0B
#define	kmNKey			0x2D
#define	kmMKey			0x2E
#define	kmSpaceKey		0x31

#define kIconSize		16
#define kKeySize		24
#define KS				kKeySize

UInt8	keyCode[kKeys] =
{
	kmUpArrowKey,
	kmDownArrowKey,
	kmLeftArrowKey,
	kmRightArrowKey,
	kmShiftKey,
	kmOptionKey,
	kmControlKey,
	kmCommandKey,
	kmZKey,
	kmXKey,
	kmReturnKey,
	kmTabKey,
	
	kmKP8Key,
	kmKP2Key,
	kmKP4Key,
	kmKP6Key,
	kmPageDownKey,
	kmPageUpKey,
	kmEndKey,
	kmHomeKey,
	kmKP0Key,
	kmKPPeriodKey,
	kmKPEnterKey,
	kmKPPlusKey,
	
	kmBackslashKey,
	km1Key,
	km0Key,
	kmTildeKey,
	kmRKey,
	
	kmBKey,
	kmNKey,
	kmMKey,
	
	kmSpaceKey
};

typedef struct
{ 
	uint8	keyWidth, keyHeight, scancode;
  	char	*keyLabel;
}	KeyboardLayout;

static void PlotKeyIcon(short);
static void UnplotKeyIcon(short);
static void RenderKey(KeyboardLayout *, Rect *);
static short FindHitKey(Point);
static short DragGrayRect(Point, short);
static Boolean KeyCodeInUse(short);
static pascal OSStatus KeyboardEventHandler(EventHandlerCallRef, EventRef, void *);

static GWorldPtr	keyboardWorld, keyBuffer, keyDrawBuffer, keyBackBuffer;
static IconSuiteRef	keyIcon[kKeys];
static Rect			keyboardRect;
static Rect			keyRect[0x80][2];
static UInt8		defaultKeys[kKeys];
static WindowRef	keyWRef;

static const Rect	keyboardSize = { 0, 0, kKeySize * 7 + 1, kKeySize * 23 + 1 };

static KeyboardLayout	keys[] =
{
	{ KS, 		KS, 	0x35, 	"esc"    },
	{ KS, 		KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x7a, 	"F1"     },
	{ KS, 		KS, 	0x78, 	"F2"     },
	{ KS, 		KS, 	0x63, 	"F3"     },
	{ KS, 		KS, 	0x76, 	"F4"     },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x60, 	"F5"     },
	{ KS, 		KS, 	0x61, 	"F6"     },
	{ KS, 		KS, 	0x62, 	"F7"     },
	{ KS, 		KS, 	0x64, 	"F8"     },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x65, 	"F9"     },
	{ KS, 		KS, 	0x6d, 	"F10"    },
	{ KS, 		KS, 	0x67, 	"F11"    },
	{ KS, 		KS, 	0x6f, 	"F12"    },
	{ KS/2, 	KS, 	0x00,	nil      },
	{ KS, 		KS, 	0x69, 	"F13"    },
	{ KS, 		KS, 	0x6b, 	"F14"    },
	{ KS, 		KS, 	0x71, 	"F15"    },
	{ 0,  		0,  	0x00, 	nil      },

	{ 0,  		0,  	0x00, 	nil      },

	{ KS, 		KS, 	0x32, 	"`"      },
	{ KS, 		KS, 	0x12, 	"1"      },
	{ KS, 		KS, 	0x13, 	"2"      },
	{ KS, 		KS, 	0x14, 	"3"      },
	{ KS, 		KS, 	0x15, 	"4"      },
	{ KS, 		KS, 	0x17, 	"5"      },
	{ KS, 		KS, 	0x16, 	"6"      },
	{ KS, 		KS, 	0x1a, 	"7"      },
	{ KS, 		KS, 	0x1c, 	"8"      },
	{ KS, 		KS, 	0x19, 	"9"      },
	{ KS, 		KS, 	0x1d, 	"0"      },
	{ KS, 		KS, 	0x1b, 	"-"      },
	{ KS, 		KS, 	0x18, 	"="      },
	{ KS*2, 	KS, 	0x33, 	"delete" },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x72, 	"ins"    },
	{ KS, 		KS, 	0x73, 	"hom"    },
	{ KS, 		KS, 	0x74, 	"pgu"    },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x47, 	"clr"    },
	{ KS, 		KS, 	0x51, 	"="      },
	{ KS, 		KS, 	0x4b, 	"/"      },
	{ KS, 		KS, 	0x43, 	"*"      },
	{ 0,  		0,  	0x00, 	nil      },

	{ KS*3/2, 	KS, 	0x30, 	"tab"    },
	{ KS, 		KS, 	0x0c, 	"Q"      },
	{ KS, 		KS, 	0x0d, 	"W"      },
	{ KS, 		KS, 	0x0e, 	"E"      },
	{ KS, 		KS, 	0x0f, 	"R"      },
	{ KS, 		KS, 	0x11, 	"T"      },
	{ KS, 		KS, 	0x10, 	"Y"      },
	{ KS, 		KS, 	0x20, 	"U"      },
	{ KS, 		KS, 	0x22, 	"I"      },
	{ KS, 		KS, 	0x1f, 	"O"      },
	{ KS, 		KS, 	0x23, 	"P"      },
	{ KS, 		KS, 	0x21, 	"["      },
	{ KS, 		KS, 	0x1e, 	"]"      },
	{ KS*3/2, 	KS, 	0x2a, 	"\\"     },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x75, 	"del"    },
	{ KS, 		KS, 	0x77, 	"end"    },
	{ KS, 		KS, 	0x79, 	"pgd"    },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x59, 	"7"      },
	{ KS, 		KS, 	0x5b, 	"8"      },
	{ KS, 		KS, 	0x5c, 	"9"      },
	{ KS, 		KS, 	0x4e, 	"-"      },
	{ 0,  		0,  	0x00, 	nil      },
	
	{ KS*2, 	KS, 	0x39, 	"caps"   },
	{ KS, 		KS, 	0x00, 	"A"      },
	{ KS, 		KS, 	0x01, 	"S"      },
	{ KS, 		KS, 	0x02, 	"D"      },
	{ KS, 		KS, 	0x03, 	"F"      },
	{ KS, 		KS, 	0x05, 	"G"      },
	{ KS, 		KS, 	0x04, 	"H"      },
	{ KS, 		KS, 	0x26, 	"J"      },
	{ KS, 		KS, 	0x28, 	"K"      },
	{ KS, 		KS, 	0x25, 	"L"      },
	{ KS, 		KS, 	0x29, 	";"      },
	{ KS, 		KS, 	0x27, 	""      },
	{ KS*2, 	KS, 	0x24, 	"return" },
	{ KS*4, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x56, 	"4"      },
	{ KS, 		KS, 	0x57, 	"5"      },
	{ KS, 		KS, 	0x58, 	"6"      },
	{ KS, 		KS, 	0x45, 	"+"      },
	{ 0,  		0,  	0x00, 	nil      },
	
	{ KS*5/2, 	KS, 	0x38, 	"shift"  },
	{ KS, 		KS, 	0x06, 	"Z"      },
	{ KS, 		KS, 	0x07,	"X"      },
	{ KS, 		KS, 	0x08, 	"C"      },
	{ KS, 		KS, 	0x09, 	"V"      },
	{ KS, 		KS, 	0x0b, 	"B"      },
	{ KS, 		KS, 	0x2d, 	"N"      },
	{ KS, 		KS, 	0x2e, 	"M"      },
	{ KS, 		KS, 	0x2b, 	","      },
	{ KS, 		KS, 	0x2f, 	"."      },
	{ KS, 		KS, 	0x2c, 	"/"      },
	{ KS*5/2, 	KS, 	0x38, 	"shift"  },
	{ KS*3/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x7e, 	"up"     },
	{ KS*3/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x53, 	"1"      },
	{ KS, 		KS, 	0x54, 	"2"      },
	{ KS, 		KS, 	0x55, 	"3"      },
	{ KS, 		KS*2, 	0x4c, 	"ent"    },
	{ 0,  		0,  	0x00, 	nil      },
	
	{ KS*3/2, 	KS, 	0x3b, 	"ctrl"   },
	{ KS*3/2, 	KS, 	0x3a, 	"opt"    },
	{ KS*3/2, 	KS, 	0x37, 	"cmd"    },
	{ KS*6, 	KS, 	0x31, 	"   "    },
	{ KS*3/2, 	KS, 	0x37, 	"cmd"    },
	{ KS*3/2, 	KS, 	0x3a, 	"opt"    },
	{ KS*3/2, 	KS, 	0x3b, 	"ctrl"   },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS, 		KS, 	0x7b, 	"lt"     },
	{ KS, 		KS, 	0x7d, 	"dn"     },
	{ KS, 		KS, 	0x7c, 	"rt"     },
	{ KS/2, 	KS, 	0x00, 	nil      },
	{ KS*2, 	KS, 	0x52, 	"0"      },
	{ KS, 		KS, 	0x41, 	"."      },
	{ 0,  		0,  	0x00, 	nil      },
};

void ConfigureKeyboard(void)
{
	OSStatus		err;
	IBNibRef		nibRef;
	CGrafPtr		oldport;
	Rect			dialogRect;
	int				i;
	EventHandlerRef	eref;
	EventHandlerUPP	eventUPP;
	EventTypeSpec	windowEvents[] = { { kEventClassWindow,  kEventWindowClose       },
									   { kEventClassWindow,  kEventWindowDrawContent },
									   { kEventClassMouse,   kEventMouseDown         },
									   { kEventClassCommand, kEventCommandProcess    } };
	
	const Rect		keyBufferRect  = { 0, 0, kIconSize + 2, kIconSize + 2 };

	err = CreateNibReference(CFSTR("SNES9X"), &nibRef);
	if (err)
		return;
	
	err = CreateWindowFromNib(nibRef, CFSTR("KeyConfig"), &keyWRef);
	if (err)
		return;
		
	for (i = 0; i < kKeys; i++)
		GetIconSuite(&keyIcon[i], 800 + i, kSelectorSmall8Bit | kSelectorSmall1Bit);
	
	GetPort(&oldport);
	SetPortWindowPort(keyWRef);
	
	SetRect(&keyboardRect, 20, 20, 573, 191);
	keyboardRect.bottom = keyboardRect.top  + keyboardSize.bottom;
	keyboardRect.right  = keyboardRect.left + keyboardSize.right;

	GetWindowPortBounds(keyWRef, &dialogRect);
	InitGWorld(&keyBuffer,     &keyBufferRect, 32);
	InitGWorld(&keyDrawBuffer, &keyBufferRect, 32);
	InitGWorld(&keyBackBuffer, &dialogRect,    32);
	
	err = InstallStandardEventHandler(GetWindowEventTarget(keyWRef));
	eventUPP = NewEventHandlerUPP(KeyboardEventHandler);
	err = InstallWindowEventHandler(keyWRef, eventUPP, 4, windowEvents, (void *) keyWRef, &eref);

	MoveWindowPosition(keyWRef, kWindowKeyConfig);
	ShowWindow(keyWRef);
	err = RunAppModalLoopForWindow(keyWRef);
	HideWindow(keyWRef);
	SaveWindowPosition(keyWRef, kWindowKeyConfig);
	
	err = RemoveEventHandler(eref);
	DisposeEventHandlerUPP(eventUPP);
	
	DisposeGWorld(keyBuffer);
	DisposeGWorld(keyDrawBuffer);
	DisposeGWorld(keyBackBuffer);
	
	DisposeWindow(keyWRef);
	DisposeNibReference(nibRef);

	SetPort(oldport);
			
	for (i = 0; i < kKeys; i++)
		if (keyIcon[i])
			DisposeIconSuite(keyIcon[i], true);	
}

void DeinitKeyboard(void)
{
	if (keyboardWorld)
		DisposeGWorld(keyboardWorld);
}

void InitKeyboard(void)
{
	OSStatus			err;
	ThemeDrawingState	state;
	Rect				myRect;
	short				row, index, scancode;
	
	memcpy(defaultKeys, keyCode, sizeof(keyCode));
	
	for (index = 0; index < 0x80; index++)
	{
		SetRect(&(keyRect[index][0]), 0, 0, 0, 0);
		SetRect(&(keyRect[index][1]), 0, 0, 0, 0);
	}

	InitGWorld(&keyboardWorld, &keyboardSize, 32);
	PrepareForGDrawing(keyboardWorld);

	err = GetThemeDrawingState(&state);

	SetThemePen(kThemeBrushMovableModalBackground, 32, true);
	PaintRect(&keyboardSize);
	TextFont(kFontIDGeneva);
	TextSize(10);
	
	SetRect(&myRect, 0, 0, 0, 0);	
	index = 0;

	for (row = 0; row < 7; row++)
	{	
		while (keys[index].keyWidth)
		{
			myRect.right += keys[index].keyWidth;
			
			if (keys[index].keyLabel)
			{
				myRect.bottom = myRect.top + keys[index].keyHeight + 1;
				scancode = keys[index].scancode;
				
				if (keyRect[scancode][0].bottom == 0)
					keyRect[scancode][0] = myRect;
				else
					keyRect[scancode][1] = myRect;
				
				RenderKey(&keys[index], &myRect);
			}
			
			myRect.left = myRect.right;
			index++;
		}
		
		OffsetRect(&myRect, 0, kKeySize);
		myRect.left = myRect.right = 0;
		index++;
	}
	
	err = SetThemeDrawingState(state, true);

	FinishGDrawing(keyboardWorld);
}

static void RenderKey(KeyboardLayout *key, Rect *rect)
{
	RGBColor			foreColor;
	short				textInset;
	ThemeButtonDrawInfo	info = { kThemeStateActive, kThemeButtonOff, kThemeAdornmentNone };
	
	rect->bottom--;
	
	DrawThemeButton(rect, kThemeSmallBevelButton, &info, nil, nil, nil, 0L);
	
	rect->bottom++;

	rect->right++;
	
	GetForeColor(&foreColor);
	
	textInset = ((key->scancode >= 0x67 && key->scancode <= 0x71) || (key->scancode == 0x73)) ? 2 : 4;	
	MoveTo(rect->left + textInset + 1, rect->bottom - 6);
	ForeColor(whiteColor);
	DrawText(key->keyLabel, 0, strlen(key->keyLabel));
	MoveTo(rect->left + textInset, rect->bottom - 7);
	ForeColor(blackColor);
	DrawText(key->keyLabel, 0, strlen(key->keyLabel));
	
	RGBForeColor(&foreColor);
	
	rect->right--;
}

static void PlotKeyIcon(short which)
{
	RGBColor	foreColor, backColor;
	CGrafPtr	oldPort;
	Rect		worldRect, myRect, tempRect;
	short		each;

	const RGBColor	grayColor = { 0xA000, 0xA000, 0xC000 };
	
	GetPort(&oldPort);
	SetPortWindowPort(keyWRef);
	
	GetForeColor(&foreColor);
	GetBackColor(&backColor);
	
	ForeColor(blackColor);
	BackColor(whiteColor);
	
	for (each = 0; each <= 1; each++)
	{
		worldRect = myRect = keyRect[keyCode[which]][each];
		
		if (myRect.bottom > 0)
		{
			OffsetRect(&myRect, keyboardRect.left, keyboardRect.top);
			tempRect = myRect;
			InsetRect(&tempRect, 1, 1);
			tempRect.bottom--;
			PaintRect(&tempRect);
			OpColor(&grayColor);
			CopyBits(GetPortBitMapForCopyBits(keyboardWorld), GetPortBitMapForCopyBits(GetWindowPort(keyWRef)), &worldRect, &myRect, blend, nil);

			myRect.left += (myRect.right  - myRect.left - kIconSize) >> 1;
			myRect.top  += (myRect.bottom - myRect.top  - kIconSize) >> 1;
			myRect.right  = myRect.left + kIconSize;
			myRect.bottom = myRect.top  + kIconSize;
			
			PlotIconSuite(&myRect, kAlignAbsoluteCenter, kTransformNone, keyIcon[which]);
	
			InsetRect(&myRect, -1, -1);
			FrameRect(&myRect);
		}
	}

	RGBForeColor(&foreColor);
	RGBBackColor(&backColor);
	
	SetPort(oldPort);
}

static void UnplotKeyIcon(short which)
{
	RGBColor	foreColor, backColor;
	CGrafPtr	oldPort;
	Rect		worldRect, myRect;
	short		each;
	
	GetPort(&oldPort);
	SetPortWindowPort(keyWRef);

	GetForeColor(&foreColor);
	GetBackColor(&backColor);
	
	ForeColor(blackColor);
	BackColor(whiteColor);

	for (each = 0; each <= 1; each++)
	{
		worldRect = myRect = keyRect[keyCode[which]][each];
		
		if (myRect.bottom > 0)
		{
			OffsetRect(&myRect, keyboardRect.left, keyboardRect.top);
			CopyBits(GetPortBitMapForCopyBits(keyboardWorld), GetPortBitMapForCopyBits(GetWindowPort(keyWRef)), &worldRect, &myRect, srcCopy, nil);
		}
	}
	
	RGBForeColor(&foreColor);
	RGBBackColor(&backColor);
	
	SetPort(oldPort);
}

static short DragGrayRect(Point mouseAt, short changeKey)
{
	CGrafPtr			port, oldPort;
	RGBColor			foreColor, backColor; 
	Rect				oldRect, drawRect, portRect;
	Point				newMouse;
	MouseTrackingResult	trackingResult;
	short				which, each;
		
	const RGBColor	grayColor  = { 0xA000, 0xA000, 0xC000 };
	const Rect		bufferRect = { 0, 0, kIconSize + 2, kIconSize + 2 };
	
	port = GetWindowPort(keyWRef);
	GetPort(&oldPort);
	SetPort(port);
	
	GetForeColor(&foreColor);
	GetBackColor(&backColor);

	newMouse.h = mouseAt.h - keyboardRect.left;
	newMouse.v = mouseAt.v - keyboardRect.top;
	
	if (PtInRect(newMouse, &keyRect[keyCode[changeKey]][0]))
		drawRect = keyRect[keyCode[changeKey]][0];
	else
		drawRect = keyRect[keyCode[changeKey]][1];
	
	drawRect.left  += (drawRect.right  - drawRect.left - kIconSize) >> 1;
	drawRect.top   += (drawRect.bottom - drawRect.top  - kIconSize) >> 1;
	drawRect.right  =  drawRect.left + kIconSize;
	drawRect.bottom =  drawRect.top  + kIconSize;

	OffsetRect(&drawRect, keyboardRect.left, keyboardRect.top);	
	InsetRect(&drawRect, -1, -1);
	
	ForeColor(blackColor);
	BackColor(whiteColor);

	CopyBits(GetPortBitMapForCopyBits(port), GetPortBitMapForCopyBits(keyBuffer),     &drawRect, &bufferRect, srcCopy, nil);
	
	GetPortBounds(port, &portRect);
	CopyBits(GetPortBitMapForCopyBits(port), GetPortBitMapForCopyBits(keyBackBuffer), &portRect, &portRect,   srcCopy, nil);

	PrepareForGDrawing(keyDrawBuffer);
	OpColor(&grayColor);
	FinishGDrawing(keyDrawBuffer);
	
	SetPort(port);	

	trackingResult = kMouseTrackingMouseDown;
	while (trackingResult != kMouseTrackingMouseUp)
	{
		TrackMouseLocation(nil, &newMouse, &trackingResult);
		if (newMouse.h != mouseAt.h || newMouse.v != mouseAt.v)
		{
			oldRect = drawRect;
			OffsetRect(&drawRect, newMouse.h - mouseAt.h, newMouse.v - mouseAt.v);
			mouseAt = newMouse;
						
			PrepareForGDrawing(keyDrawBuffer);
			CopyBits(GetPortBitMapForCopyBits(keyBackBuffer), GetPortBitMapForCopyBits(keyDrawBuffer), &drawRect,   &bufferRect, srcCopy, nil);
			CopyBits(GetPortBitMapForCopyBits(keyBuffer),     GetPortBitMapForCopyBits(keyDrawBuffer), &bufferRect, &bufferRect, blend,   nil);
			FinishGDrawing(keyDrawBuffer); 
			
			SetPort(port);

			CopyBits(GetPortBitMapForCopyBits(keyBackBuffer), GetPortBitMapForCopyBits(port), &oldRect,    &oldRect,  srcCopy, nil);
			CopyBits(GetPortBitMapForCopyBits(keyDrawBuffer), GetPortBitMapForCopyBits(port), &bufferRect, &drawRect, srcCopy, nil);
		}
	}

	CopyBits(GetPortBitMapForCopyBits(keyBackBuffer), GetPortBitMapForCopyBits(port), &drawRect, &drawRect, srcCopy, nil);

	RGBForeColor(&foreColor);
	RGBBackColor(&backColor);
	
	SetPort(oldPort);

	mouseAt.h -= keyboardRect.left;
	mouseAt.v -= keyboardRect.top;
	
	for (which = 0; which < 0x80; which++)
	{
		for (each = 0; each <= 1; each++)
		{
			if (PtInRect(mouseAt, &keyRect[which][each]))
			{
				if (!KeyCodeInUse(which))
					return which;
			}
		}
	}
	
	return -1;
}

static Boolean KeyCodeInUse(short which)
{
	short	count;
	
	for (count = 0; count < kKeys; count++)
		if (keyCode[count] == which)
			return true;

	return false;
}

static short FindHitKey(Point where)
{
	short	hit, which, each;
	
	where.v -= keyboardRect.top;
	where.h -= keyboardRect.left;

	hit = -1;
	for (which = 0; which < kKeys; which++)
		for (each = 0; each <= 1; each++)
		{
			if (PtInRect(where, &keyRect[keyCode[which]][each]))
				hit = which;
		};
	
	return hit;
}

static pascal OSStatus KeyboardEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
	#pragma unused (inHandlerRef)
	
	OSStatus	err, result = eventNotHandledErr;
	WindowRef	tWindowRef = (WindowRef) inUserData;

	switch (GetEventClass(inEvent))
	{
		case kEventClassWindow:
			switch (GetEventKind(inEvent))
			{
				case kEventWindowClose:
					QuitAppModalLoopForWindow(tWindowRef);
					result = noErr;
					
					break;
					
				case kEventWindowDrawContent:
					RGBColor	foreColor, backColor;
					Rect		portRect;
			
					GetForeColor(&foreColor);
					GetBackColor(&backColor);
					
					ForeColor(blackColor);
					BackColor(whiteColor);
					GetPortBounds(keyboardWorld, &portRect);
					CopyBits(GetPortBitMapForCopyBits(keyboardWorld), GetPortBitMapForCopyBits(GetWindowPort(tWindowRef)), &portRect, &keyboardRect, srcCopy, nil);				

					for (int which = 0; which < kKeys; which++)
						PlotKeyIcon(which);
					
					RGBForeColor(&foreColor);
					RGBBackColor(&backColor);
					
					result = noErr;
			}
			
			break;
			
		case kEventClassMouse:
			switch (GetEventKind(inEvent))
			{
				case kEventMouseDown:
					Point	localHit;
					short	changeKey, changeTo;
					
					err = GetEventParameter(inEvent, kEventParamMouseLocation, typeQDPoint, nil, sizeof(Point), nil, &localHit);
					if (err == noErr)
					{
						GlobalToLocal(&localHit);
			
						if (PtInRect(localHit, &keyboardRect))
						{
							changeKey = FindHitKey(localHit);
							if (changeKey != -1)
							{
								changeTo = DragGrayRect(localHit, changeKey);
								
								if (changeTo != -1)
								{
									UnplotKeyIcon(changeKey);
									keyCode[changeKey] = changeTo;
									PlotKeyIcon(changeKey);
								}
							}
													
							result = noErr;
						}
					}
			}
			
			break;
			
		case kEventClassCommand:
			switch (GetEventKind(inEvent))
			{
				case kEventCommandProcess:
					HICommand	tHICommand;
					
					err = GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, nil, sizeof(HICommand), nil, &tHICommand);
					if (err == noErr)
					{
						if (tHICommand.commandID == 'DFLT')
						{
							memcpy(keyCode, defaultKeys, sizeof(keyCode));
							InvalWindowRect(tWindowRef, &keyboardRect);
							
							result = noErr;
						}
					}
			}
	}
	
	return result;
}
