/*******************************************************************************
  Snes9x - Portable Super Nintendo Entertainment System (TM) emulator.
 
  (c) Copyright 1996 - 2003 Gary Henderson (gary.henderson@ntlworld.com) and
                            Jerremy Koot (jkoot@snes9x.com)

  (c) Copyright 2002 - 2003 Matthew Kendora and
                            Brad Jorsch (anomie@users.sourceforge.net)
 

                      
  C4 x86 assembler and some C emulation code
  (c) Copyright 2000 - 2003 zsKnight (zsknight@zsnes.com),
                            _Demo_ (_demo_@zsnes.com), and
                            Nach (n-a-c-h@users.sourceforge.net)
                                          
  C4 C++ code
  (c) Copyright 2003 Brad Jorsch

  DSP-1 emulator code
  (c) Copyright 1998 - 2003 Ivar (ivar@snes9x.com), _Demo_, Gary Henderson,
                            John Weidman (jweidman@slip.net),
                            neviksti (neviksti@hotmail.com), and
                            Kris Bleakley (stinkfish@bigpond.com)
 
  DSP-2 emulator code
  (c) Copyright 2003 Kris Bleakley, John Weidman, neviksti, Matthew Kendora, and
                     Lord Nightmare (lord_nightmare@users.sourceforge.net

  OBC1 emulator code
  (c) Copyright 2001 - 2003 zsKnight, pagefault (pagefault@zsnes.com)
  Ported from x86 assembler to C by sanmaiwashi

  SPC7110 and RTC C++ emulator code
  (c) Copyright 2002 Matthew Kendora with research by
                     zsKnight, John Weidman, and Dark Force

  S-RTC C emulator code
  (c) Copyright 2001 John Weidman
  
  Super FX x86 assembler emulator code 
  (c) Copyright 1998 - 2003 zsKnight, _Demo_, and pagefault 

  Super FX C emulator code 
  (c) Copyright 1997 - 1999 Ivar and Gary Henderson.



 
  Specific ports contains the works of other authors. See headers in
  individual files.
 
  Snes9x homepage: http://www.snes9x.com
 
  Permission to use, copy, modify and distribute Snes9x in both binary and
  source form, for non-commercial purposes, is hereby granted without fee,
  providing that this license information and copyright notice appear with
  all copies and any derived work.
 
  This software is provided 'as-is', without any express or implied
  warranty. In no event shall the authors be held liable for any damages
  arising from the use of this software.
 
  Snes9x is freeware for PERSONAL USE only. Commercial users should
  seek permission of the copyright holders first. Commercial use includes
  charging money for Snes9x or software derived from Snes9x.
 
  The copyright holders request that bug fixes and improvements to the code
  should be forwarded to them so everyone can benefit from the modifications
  in future versions.
 
  Super NES and Super Nintendo Entertainment System are trademarks of
  Nintendo Co., Limited and its subsidiary companies.
*******************************************************************************/

/*******************************************************************************
  Snes9x custom (Snes9x for Mac OS) by zones
  
  This build was started from John Stiles' port.
*******************************************************************************/

#include "gfx.h"
#include "spc7110.h"

#include "mac-prefix.h"
#include "mac-audio.h"
#include "mac-dialog.h"
#include "mac-keyboard.h"
#include "mac-os.h"
#include "mac-snes9x.h"
#include "mac-stringtools.h"
#include "mac-prefs.h"

static void SelectItemOfTabControl(ControlRef, SInt16);
static pascal OSStatus TabEventHandler(EventHandlerCallRef, EventRef, void *);
static pascal OSStatus PreferencesEventHandler(EventHandlerCallRef, EventRef, void *);

static int	lastTabIndex = 1;
static int	tabList[]    = { 5, 129, 130, 131, 132, 133 };

struct PrefList
{
	long	itemName;
	void	*itemPointer;
	short	size;
};

static PrefList	prefList[] =
{
	{ 'flsc', &fullscreen,                          sizeof(bool8      ) },
	{ 'reso', &autoRes,                             sizeof(bool8      ) },
	{ 'dfps', &Settings.DisplayFrameRate,           sizeof(bool8      ) },
	{ '7int', &Settings.Mode7Interpolate,           sizeof(bool8      ) },
	{ 'tran', &Settings.Transparency,               sizeof(bool8      ) },
	{ 'gl32', &gl32bit,                             sizeof(bool8      ) },
	{ 'glst', &glstretch,                           sizeof(bool8      ) },
	{ 'glpr', &glproportional,                      sizeof(bool8      ) },
	{ 'draw', &drawingMethod,                       sizeof(long       ) },
	{ '2xgr', &doubleSize,                          sizeof(bool8      ) },
	{ 't.v.', &tvMode,                              sizeof(bool8      ) },
	{ 'fltr', &smoothMode,                          sizeof(bool8      ) },
	{ 'sai ', &saiMode,                             sizeof(bool8      ) },
	{ 'sSai', &supsaiMode,                          sizeof(bool8      ) },
	{ 'eagl', &eagleMode,                           sizeof(bool8      ) },
	{ 'MPmt', &multiprocessor,                      sizeof(bool8      ) },
	{ 'VSNC', &vsync,                               sizeof(bool8      ) },
	{ 'H239', &drawoverscan,						sizeof(bool8      ) },
	
	{ 'audi', &Settings.APUEnabled,                 sizeof(bool8      ) },
	{ 'so16', &Settings.SixteenBitSound,            sizeof(bool8      ) },
	{ 'ster', &Settings.Stereo,                     sizeof(bool8      ) },
	{ 'rbst', &Settings.ReverseStereo,              sizeof(bool8      ) },
	{ 'sint', &Settings.InterpolatedSound,          sizeof(bool8      ) },
	{ 'ehgt', &Settings.SoundEnvelopeHeightReading, sizeof(bool8      ) },
	{ 'echo', &Settings.DisableSoundEcho,           sizeof(bool8      ) },
//	{ 'disc', &Settings.DisableSampleCaching,       sizeof(bool8      ) },
	{ 'mstv', &Settings.DisableMasterVolume,        sizeof(bool8      ) },
	{ 'srat', &Settings.SoundPlaybackRate,          sizeof(uint32     ) },
	{ 'deco', &Settings.AltSampleDecode,            sizeof(uint8      ) },
	{ 'pich', &macSoundPitch,                       sizeof(float      ) },
	{ 'Volm', &macSoundVolume,                      sizeof(SInt32     ) },
	{ 'TMiv', &macSoundInterval,                    sizeof(int        ) },
	{ 'AUef', &aueffect,                            sizeof(uint16     ) },
	{ 'AUce', &cureffect,                           sizeof(int        ) },
		
	{ 'romf', &saveInROMFolder,                     sizeof(uint8      ) },
	{ 'atsv', &Settings.AutoSaveDelay,              sizeof(int32      ) },
	{ '7110', &mac7110Load,                         sizeof(int        ) },
	{ '711M', &mac7110Megs,                         sizeof(int        ) },
	{ 'SD1P', &macSDD1Pack,                         sizeof(int        ) },
	
	{ 'CPUc', &Settings.CyclesPercentage,           sizeof(long       ) },
	{ 'stdn', &Settings.ShutdownMaster,             sizeof(bool8      ) },
	
	{ 'StOp', &startopendlog,                       sizeof(bool8      ) },
	{ 'STiF', &showtimeinfrz,                       sizeof(bool8      ) },
	{ 'Togl', &enabletoggle,                        sizeof(bool8      ) },
	{ 'AlEv', &allowevent,                          sizeof(bool8      ) },
	{ 'SvWp', &savewindowpos,                       sizeof(bool8      ) },
	{ 'MbxM', &musicboxmode,                        sizeof(int        ) },

	{ 'tab ', &lastTabIndex,                        sizeof(int        ) },
	{ 'keyb', keyCode,                              sizeof(keyCode    ) },
	{ 'pset', &padSetting,                          sizeof(int        ) },
	{ 'dset', &deviceSetting,                       sizeof(int        ) },
	{ 'scop', &scopecursor,                         sizeof(bool8      ) },
	{ 'chea', &applycheat,                          sizeof(bool8      ) },
	{ 'swap', &Settings.SwapJoypads,                sizeof(bool8      ) },
	{ '16bt', &Settings.SixteenBit,                 sizeof(bool8      ) },
	
	{ 'ARec', autofireRec,                          sizeof(autofireRec) },
	{ 'WPos', windowPos,                 		    sizeof(windowPos  ) }
};

#define	kPrefListSize	(sizeof(prefList) / sizeof(prefList[0]))

enum
{
	iNibGFullScreen = 1,
	iNibGSwitchResolution,
	iNibGShowFrameRate,
	iNibGMode7Smoothing,
	iNibGTransparency,
	iNibGGL32bit,
	iNibGGLStretch,
	iNibGGLProportional,
	iNibGVideoMode,
	iNibGDirectMP,
	iNibGGLVSync,
	iNibGDrawOverscan,
	
	iNibSEnableSound = 201,
	iNibS16BitPlayback,
	iNibSStereo,
	iNibSReverseStereo,
	iNibSInterpolation,
	iNibSHeightReading,
	iNibSEcho,
	iNibSCacheSample,   // unused
	iNibSMasterVolume,
	iNibSPlaybackRate,
	iNibSSampleDecoder,
	iNibSReserved1,		// unused
	iNibSReserved2,		// unused
	iNibSPitch,
	iNibSVolume,
	iNibSMixInterval,
	
	iNibOSaveFolder = 401,
	iNibOReserved1,		// unused
	iNibOReserved2,		// unused
	iNibOAutoSaveInterval,
	iNibOSPC7110Pack,
	iNibOSPC7110Megs,
	iNibOSDD1Pack,
	
	iNibMCPUCycles = 601,
	iNibMShutdownMaster,
	
	iNibXStartOpenDialog = 801,
	iNibXShowTimeInFrz,
	iNibXMusicBoxMode,
	iNibXEnableToggle,
	iNibXAllowEvent,
	iNibXSaveWindowPos,
	
	iDirectSmall = 1,
	iDirectBlocky,
	iDirectTVMode,
	iDirectSmoothMode,
	iDirectEAGLEMode,
	iDirect2xSAIMode,
	iDirectSuper2xSAIMode,
	iOpenGLBlocky = 9,
	iOpenGLTVMode,
	iOpenGLSmoothMode,
	iOpenGLEagleMode,
	iOpenGL2xSAIMode,
	iOpenGLSuper2xSAIMode,
	
	iBufferSmall = 1,
	iBufferLarge = 2
};

void SavePrefs(void)
{
	CFMutableStringRef	mref;
	CFStringRef			sref;
	CFDataRef			data;
	
	for (unsigned int i = 0; i < kPrefListSize; i++)
	{		
		mref = CFStringCreateMutableCopy(kCFAllocatorDefault, 0, CFSTR("Preferences_"));
		if (mref)
		{
			sref = CFStringCreateWithBytes(kCFAllocatorDefault, (UInt8 *) &(prefList[i].itemName), sizeof(long), kCFStringEncodingMacRoman, false);
			if (sref)
			{
				CFStringAppend(mref, sref);

				data = CFDataCreate(kCFAllocatorDefault, (UInt8 *) prefList[i].itemPointer, prefList[i].size);
				if (data)
				{
					CFPreferencesSetAppValue(mref, data, kCFPreferencesCurrentApplication);
					CFRelease(data);
				}

				CFRelease(sref);
			}

			CFRelease(mref);
		}
	}
	
	CFPreferencesAppSynchronize(kCFPreferencesCurrentApplication);
}

void LoadPrefs(void)
{
	CFMutableStringRef	mref;
	CFStringRef			sref;
	CFDataRef			data;
	
	for (unsigned int i = 0; i < kPrefListSize; i++)
	{		
		mref = CFStringCreateMutableCopy(kCFAllocatorDefault, 0, CFSTR("Preferences_"));
		if (mref)
		{
			sref = CFStringCreateWithBytes(kCFAllocatorDefault, (UInt8 *) &(prefList[i].itemName), sizeof(long), kCFStringEncodingMacRoman, false);
			if (sref)
			{
				CFStringAppend(mref, sref);
				
				data = CFPreferencesCopyAppValue(mref, kCFPreferencesCurrentApplication);
				if (data)
				{
					if (CFDataGetLength(data) == prefList[i].size)
						CFDataGetBytes(data, CFRangeMake(0, prefList[i].size), (UInt8 *) prefList[i].itemPointer);
					CFRelease(data);
				}
				
				CFRelease(sref);
			}
			
			CFRelease(mref);
		}
	}
}

void ConfigurePreferences(void)
{
	OSStatus	err;
	IBNibRef	nibRef;
	
	err = CreateNibReference(CFSTR("SNES9X"), &nibRef);
	if (noErr == err)
	{
		WindowRef	tWindowRef;
		uint32 		oldPlaybackRate;
		bool8		oldEnableSound, old16BitPlayback, oldStereo, oldFullscreen;
		int			oldMixInterval;
		SInt32		oldVolume;
		
		oldEnableSound   = Settings.APUEnabled;
		old16BitPlayback = Settings.SixteenBitSound;
		oldStereo        = Settings.Stereo;
		oldPlaybackRate  = Settings.SoundPlaybackRate;
		oldMixInterval   = macSoundInterval;
		oldVolume        = macSoundVolume;
		oldFullscreen    = fullscreen;

		S9xGraphicsDeinit();
		
		err = CreateWindowFromNib(nibRef, CFSTR("Preferences"), &tWindowRef);
		if (noErr == err)
		{
			ControlRef		control;
			ControlID		controlID;
			EventHandlerUPP	tabEventProc, prefEventProc;
			EventHandlerRef	tabHandlerRef, prefHandlerRef;
			EventTypeSpec	tabControlSpec[]  = { { kEventClassControl, kEventControlHit  } },
							prefControlSpec[] = { { kEventClassWindow,  kEventWindowClose }, { kEventClassCommand, kEventCommandProcess } };
			MenuHandle		menu;
			char			num[16];
			unsigned char	numP[16];
			int				i;

			controlID.signature = 'tabs';
			controlID.id = 128;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, lastTabIndex);
			SelectItemOfTabControl(control, lastTabIndex);
			tabEventProc = NewEventHandlerUPP(TabEventHandler);
			err = InstallControlEventHandler(control, tabEventProc, 1, tabControlSpec, 0, &tabHandlerRef);
			
			prefEventProc = NewEventHandlerUPP(PreferencesEventHandler);
			err = InstallWindowEventHandler(tWindowRef, prefEventProc, 2, prefControlSpec, (void *) tWindowRef, &prefHandlerRef);
			
			if (!useV2)
			{
				controlID.signature = 'S_EF';
				controlID.id = 0;			
				GetControlByID(tWindowRef, &controlID, &control);
				DeactivateControl(control);
			}

			controlID.signature = 'grap';

			controlID.id = iNibGFullScreen;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, fullscreen);
			
			controlID.id = iNibGSwitchResolution;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, autoRes);
			
			controlID.id = iNibGShowFrameRate;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.DisplayFrameRate);
			
			controlID.id = iNibGMode7Smoothing;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.Mode7Interpolate);
			
			controlID.id = iNibGTransparency;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.Transparency);
			
			controlID.id = iNibGGL32bit;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, gl32bit);
			
			controlID.id = iNibGGLStretch;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, glstretch);
			
			controlID.id = iNibGGLProportional;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, glproportional);
			
			controlID.id = iNibGVideoMode;
			GetControlByID(tWindowRef, &controlID, &control);
			menu = GetControlPopupMenuHandle(control);
			for (i = 1; i <= CountMenuItems(menu); i++)
				CheckMenuItem(menu, i, false);
			switch (drawingMethod)
			{
				case kDrawingDirect:
					if (smoothMode)
					{
						CheckMenuItem(menu, iDirectSmoothMode, true);
						SetControl32BitValue(control, iDirectSmoothMode);
					}
					else
					if (eagleMode)
					{
						CheckMenuItem(menu, iDirectEAGLEMode, true);
						SetControl32BitValue(control, iDirectEAGLEMode);
					}
					else
					if (supsaiMode)
					{
						CheckMenuItem(menu, iDirectSuper2xSAIMode, true);
						SetControl32BitValue(control, iDirectSuper2xSAIMode);
					}
					else
					if (saiMode)
					{
						CheckMenuItem(menu, iDirect2xSAIMode, true);
						SetControl32BitValue(control, iDirect2xSAIMode);
					}
					else
					if (tvMode)
					{
						CheckMenuItem(menu, iDirectTVMode, true);
						SetControl32BitValue(control, iDirectTVMode);
					}
					else
					if (doubleSize)
					{
						CheckMenuItem(menu, iDirectBlocky, true);
						SetControl32BitValue(control, iDirectBlocky);
					}
					else
					{
						CheckMenuItem(menu, iDirectSmall, true);
						SetControl32BitValue(control, iDirectSmall);
					}
					
					break;
				
				case kDrawingOpenGL:
					if (smoothMode)
					{
						CheckMenuItem(menu, iOpenGLSmoothMode, true);
						SetControl32BitValue(control, iOpenGLSmoothMode);
					}
					else
					{
						CheckMenuItem(menu, iOpenGLBlocky, true);
						SetControl32BitValue(control, iOpenGLBlocky);
					}
					
					break;
					
				case kDrawingBlitGL:
					if (eagleMode)
					{
						CheckMenuItem(menu, iOpenGLEagleMode, true);
						SetControl32BitValue(control, iOpenGLEagleMode);
					}
					else
					if (supsaiMode)
					{
						CheckMenuItem(menu, iOpenGLSuper2xSAIMode, true);
						SetControl32BitValue(control, iOpenGLSuper2xSAIMode);
					}
					else
					if (saiMode)
					{
						CheckMenuItem(menu, iOpenGL2xSAIMode, true);
						SetControl32BitValue(control, iOpenGL2xSAIMode);
					}
					else
					{
						CheckMenuItem(menu, iOpenGLTVMode, true);
						SetControl32BitValue(control, iOpenGLTVMode);
					}
			}

			controlID.id = iNibGDirectMP;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, multiprocessor);
			
			controlID.id = iNibGGLVSync;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, vsync);

			controlID.id = iNibGDrawOverscan;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, drawoverscan);

			controlID.signature = 'snd_';
			
			controlID.id = iNibSEnableSound;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.APUEnabled);
			
			controlID.id = iNibS16BitPlayback;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.SixteenBitSound);
			
			controlID.id = iNibSStereo;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.Stereo);
			
			controlID.id = iNibSReverseStereo;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.ReverseStereo);
		
			controlID.id = iNibSInterpolation;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.InterpolatedSound);
			
			controlID.id = iNibSHeightReading;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.SoundEnvelopeHeightReading);

			controlID.id = iNibSEcho;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, !Settings.DisableSoundEcho);

			/*controlID.id = iNibSCacheSample;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, !Settings.DisableSampleCaching);*/

			controlID.id = iNibSMasterVolume;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, !Settings.DisableMasterVolume);
			
			controlID.id = iNibSPlaybackRate;
			GetControlByID(tWindowRef, &controlID, &control);
			menu = GetControlPopupMenuHandle(control);
			for (i = 1; i <= CountMenuItems(menu); i++)
				CheckMenuItem(menu, i, false);
			switch (Settings.SoundPlaybackRate)
			{
				case 48000:
					CheckMenuItem(menu, 1, true);
					SetControl32BitValue(control, 1);
					break;
				
				case 44100:
					CheckMenuItem(menu, 2, true);
					SetControl32BitValue(control, 2);
					break;
				
				case 35000:
					CheckMenuItem(menu, 3, true);
					SetControl32BitValue(control, 3);
					break;
				
				case 32000:
					CheckMenuItem(menu, 4, true);
					SetControl32BitValue(control, 4);
					break;
				
				case 30000:
					CheckMenuItem(menu, 5, true);
					SetControl32BitValue(control, 5);
					break;
				
				case 22050:
					CheckMenuItem(menu, 6, true);
					SetControl32BitValue(control, 6);
					break;
				
				case 16000:
					CheckMenuItem(menu, 7, true);
					SetControl32BitValue(control, 7);
					break;
					
				case 11025:
					CheckMenuItem(menu, 8, true);
					SetControl32BitValue(control, 8);
					break;
					
				case 8000:
					CheckMenuItem(menu, 9, true);
					SetControl32BitValue(control, 9);
			}

			controlID.id = iNibSSampleDecoder;
			GetControlByID(tWindowRef, &controlID, &control);		
			menu = GetControlPopupMenuHandle(control);
			for (i = 1; i <= CountMenuItems(menu); i++)
				CheckMenuItem(menu, i, false);
			switch (Settings.AltSampleDecode)
			{
				case 0:
					CheckMenuItem(menu, 1, true);
					SetControl32BitValue(control, 1);
					break;
					
				case 1:
				case 2:
					CheckMenuItem(menu, 2, true);
					SetControl32BitValue(control, 2);
					break;
					
				case 3:
					CheckMenuItem(menu, 3, true);
					SetControl32BitValue(control, 3);
			}

			controlID.id = iNibSMixInterval;
			GetControlByID(tWindowRef, &controlID, &control);
			sprintf(num, "%d", macSoundInterval);
			ConvertCString(num, numP);
			SetEditTextText(control, numP, false);

			controlID.id = iNibSPitch;
			GetControlByID(tWindowRef, &controlID, &control);
			sprintf(num, "%2.4f", macSoundPitch);
			ConvertCString(num, numP);
			SetEditTextText(control, numP, false);
			
			controlID.id = iNibSVolume;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, macSoundVolume);
			
			controlID.signature = 'othe';
			
			controlID.id = iNibOSaveFolder;
			GetControlByID(tWindowRef, &controlID, &control);		
			menu = GetControlPopupMenuHandle(control);
			for (i = 1; i <= CountMenuItems(menu); i++)
				CheckMenuItem(menu, i, false);
			if (saveInROMFolder == 1)
			{
				CheckMenuItem(menu, 2, true);
				SetControl32BitValue(control, 2);
			}
			else
			if (saveInROMFolder == 0)
			{
				CheckMenuItem(menu, 1, true);
				SetControl32BitValue(control, 1);
			}
			else
			{
				CheckMenuItem(menu, 3, true);
				SetControl32BitValue(control, 3);
			}			

			controlID.id = iNibOAutoSaveInterval;
			GetControlByID(tWindowRef, &controlID, &control);
			sprintf(num, "%ld", Settings.AutoSaveDelay);
			ConvertCString(num, numP);
			SetEditTextText(control, numP, false);

			controlID.id = iNibOSPC7110Pack;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, mac7110Load);

			controlID.id = iNibOSPC7110Megs;
			GetControlByID(tWindowRef, &controlID, &control);
			sprintf(num, "%d", mac7110Megs);
			ConvertCString(num, numP);
			SetEditTextText(control, numP, false);

			controlID.id = iNibOSDD1Pack;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, macSDD1Pack);

			controlID.signature = 'msc2';

			controlID.id = iNibMCPUCycles;
			GetControlByID(tWindowRef, &controlID, &control);
			sprintf(num, "%ld", Settings.CyclesPercentage);
			ConvertCString(num, numP);
			SetEditTextText(control, numP, false);
			
			controlID.id = iNibMShutdownMaster;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, Settings.ShutdownMaster);
			
			controlID.signature = 'osx_';
			
			controlID.id = iNibXStartOpenDialog;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, startopendlog);
			
			controlID.id = iNibXShowTimeInFrz;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, showtimeinfrz);
			
			controlID.id = iNibXMusicBoxMode;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, musicboxmode);

			controlID.id = iNibXEnableToggle;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, enabletoggle);

			controlID.id = iNibXAllowEvent;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, allowevent);

			controlID.id = iNibXSaveWindowPos;
			GetControlByID(tWindowRef, &controlID, &control);
			SetControl32BitValue(control, savewindowpos);

			MoveWindowPosition(tWindowRef, kWindowPreferences);
			ShowWindow(tWindowRef);
			err = RunAppModalLoopForWindow(tWindowRef);
			HideWindow(tWindowRef);
			SaveWindowPosition(tWindowRef, kWindowPreferences);
			
			controlID.signature = 'grap';
			
			controlID.id = iNibGFullScreen;
			GetControlByID(tWindowRef, &controlID, &control);
			fullscreen = GetControl32BitValue(control) ? true : false;
	
			controlID.id = iNibGSwitchResolution;
			GetControlByID(tWindowRef, &controlID, &control);
			autoRes = GetControl32BitValue(control) ? true : false;
		
			controlID.id = iNibGShowFrameRate;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.DisplayFrameRate = GetControl32BitValue(control) ? true : false;
		
			controlID.id = iNibGMode7Smoothing;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.Mode7Interpolate = GetControl32BitValue(control) ? true : false;
			
			controlID.id = iNibGTransparency;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.Transparency = GetControl32BitValue(control) ? true : false;
			
			controlID.id = iNibGGL32bit;
			GetControlByID(tWindowRef, &controlID, &control);
			gl32bit = GetControl32BitValue(control) ? true : false;
			
			controlID.id = iNibGGLStretch;
			GetControlByID(tWindowRef, &controlID, &control);
			glstretch = GetControl32BitValue(control) ? true : false;

			controlID.id = iNibGGLProportional;
			GetControlByID(tWindowRef, &controlID, &control);
			glproportional = GetControl32BitValue(control) ? true : false;

			controlID.id = iNibGVideoMode;
			GetControlByID(tWindowRef, &controlID, &control);
			switch (GetControl32BitValue(control))
			{
				case iDirectSmall:
					drawingMethod = kDrawingDirect;
					doubleSize    = false;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = false;
					break;

				case iDirectBlocky:
					drawingMethod = kDrawingDirect;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = false;
					break;

				case iDirectTVMode:
					drawingMethod = kDrawingDirect;
					doubleSize    = true;
					tvMode        = true;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = false;
					break;

				case iDirectSmoothMode:
					drawingMethod = kDrawingDirect;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = true;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = false;
					break;

				case iDirectEAGLEMode:
					drawingMethod = kDrawingDirect;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = true;
					saiMode       = false;
					supsaiMode    = false;
					break;
				
				case iDirect2xSAIMode:
					drawingMethod = kDrawingDirect;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = true;
					supsaiMode    = false;
					break;
					
				case iDirectSuper2xSAIMode:
					drawingMethod = kDrawingDirect;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = true;
					break;

				case iOpenGLBlocky:
					drawingMethod = kDrawingOpenGL;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = false;
					break;

				case iOpenGLSmoothMode:
					drawingMethod = kDrawingOpenGL;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = true;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = false;
					break;
					
				case iOpenGLTVMode:
					drawingMethod = kDrawingBlitGL;
					doubleSize    = true;
					tvMode        = true;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = false;
					break;

				case iOpenGLEagleMode:
					drawingMethod = kDrawingBlitGL;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = true;
					saiMode       = false;
					supsaiMode    = false;
					break;

				case iOpenGL2xSAIMode:
					drawingMethod = kDrawingBlitGL;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = true;
					supsaiMode    = false;
					break;
					
				case iOpenGLSuper2xSAIMode:
					drawingMethod = kDrawingBlitGL;
					doubleSize    = true;
					tvMode        = false;
					smoothMode    = false;
					eagleMode     = false;
					saiMode       = false;
					supsaiMode    = true;
			}
			
			controlID.id = iNibGDirectMP;
			GetControlByID(tWindowRef, &controlID, &control);
			multiprocessor = GetControl32BitValue(control) ? true : false;

			controlID.id = iNibGGLVSync;
			GetControlByID(tWindowRef, &controlID, &control);
			vsync = GetControl32BitValue(control) ? true : false;

			controlID.id = iNibGDrawOverscan;
			GetControlByID(tWindowRef, &controlID, &control);
			drawoverscan = GetControl32BitValue(control) ? true : false;

			controlID.signature = 'snd_';
			
			controlID.id = iNibSEnableSound;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.APUEnabled = Settings.NextAPUEnabled = GetControl32BitValue(control) ? true : false;
	
			controlID.id = iNibS16BitPlayback;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.SixteenBitSound = GetControl32BitValue(control) ? true : false;
			
			controlID.id = iNibSStereo;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.Stereo = GetControl32BitValue(control) ? true : false;
		
			controlID.id = iNibSReverseStereo;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.ReverseStereo = GetControl32BitValue(control) ? true : false;
		
			controlID.id = iNibSInterpolation;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.InterpolatedSound = GetControl32BitValue(control) ? true : false;

			controlID.id = iNibSHeightReading;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.SoundEnvelopeHeightReading = GetControl32BitValue(control) ? true : false;

			controlID.id = iNibSEcho;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.DisableSoundEcho = GetControl32BitValue(control) ? false : true;

			/*controlID.id = iNibSCacheSample;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.DisableSampleCaching = GetControl32BitValue(control) ? false : true;*/

			controlID.id = iNibSMasterVolume;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.DisableMasterVolume = GetControl32BitValue(control) ? false : true;

			controlID.id = iNibSPlaybackRate;
			GetControlByID(tWindowRef, &controlID, &control);
			switch (GetControl32BitValue(control))
			{
				case 1:
					Settings.SoundPlaybackRate = 48000;
					break;
					
				case 2:
					Settings.SoundPlaybackRate = 44100;
					break;
					
				case 3:
					Settings.SoundPlaybackRate = 35000;
					break;
					
				case 4:
					Settings.SoundPlaybackRate = 32000;
					break;
					
				case 5:
					Settings.SoundPlaybackRate = 30000;
					break;
					
				case 6:
					Settings.SoundPlaybackRate = 22050;
					break;
					
				case 7:
					Settings.SoundPlaybackRate = 16000;
					break;
					
				case 8:
					Settings.SoundPlaybackRate = 11025;
					break;
					
				case 9:
					Settings.SoundPlaybackRate = 8000;
			}

			controlID.id = iNibSSampleDecoder;
			GetControlByID(tWindowRef, &controlID, &control);
			switch (GetControl32BitValue(control))
			{
				case 1:
					Settings.AltSampleDecode = 0;
					break;
				
				case 2:
					Settings.AltSampleDecode = 1;
					break;
					
				case 3:
					Settings.AltSampleDecode = 3;
			}
		
			controlID.id = iNibSMixInterval;
			GetControlByID(tWindowRef, &controlID, &control);
			GetEditTextText(control, numP);
			ConvertPString(numP, num);
			macSoundInterval = atoi(num);
			if (macSoundInterval < 0)
				macSoundInterval = 0;
			if (macSoundInterval > 30)
				macSoundInterval = 30;

			controlID.id = iNibSPitch;
			GetControlByID(tWindowRef, &controlID, &control);
			GetEditTextText(control, numP);
			ConvertPString(numP, num);
			macSoundPitch = (float) atof(num);
			
			controlID.id = iNibSVolume;
			GetControlByID(tWindowRef, &controlID, &control);
			macSoundVolume = GetControl32BitValue(control);
				
			controlID.signature = 'othe';
			
			controlID.id = iNibOSaveFolder;
			GetControlByID(tWindowRef, &controlID, &control);
			int	saveto = GetControl32BitValue(control);
			if (saveto == 2)
				saveInROMFolder = 1;
			else
			if (saveto == 1)
				saveInROMFolder = 0;
			else
				saveInROMFolder = 3;

			controlID.id = iNibOAutoSaveInterval;
			GetControlByID(tWindowRef, &controlID, &control);
			GetEditTextText(control, numP);
			ConvertPString(numP, num);
			Settings.AutoSaveDelay = atoi(num);
			
			controlID.id = iNibOSPC7110Pack;
			GetControlByID(tWindowRef, &controlID, &control);
			mac7110Load = GetControl32BitValue(control);

			controlID.id = iNibOSPC7110Megs;
			GetControlByID(tWindowRef, &controlID, &control);
			GetEditTextText(control, numP);
			ConvertPString(numP, num);
			mac7110Megs = atoi(num);
			if (mac7110Megs < 1)
				mac7110Megs = 1;

			controlID.id = iNibOSDD1Pack;
			GetControlByID(tWindowRef, &controlID, &control);
			macSDD1Pack = GetControl32BitValue(control);

			controlID.signature = 'msc2';
			
			controlID.id = iNibMCPUCycles;
			GetControlByID(tWindowRef, &controlID, &control);
			GetEditTextText(control, numP);
			ConvertPString(numP, num);
			Settings.CyclesPercentage = atoi(num);
			if ((Settings.CyclesPercentage <= 0) || (Settings.CyclesPercentage >= 200))
				Settings.CyclesPercentage = 100;
				
			controlID.id = iNibMShutdownMaster;
			GetControlByID(tWindowRef, &controlID, &control);
			Settings.ShutdownMaster = GetControl32BitValue(control) ? true : false;
			
			controlID.signature = 'osx_';
			
			controlID.id = iNibXStartOpenDialog;
			GetControlByID(tWindowRef, &controlID, &control);
			startopendlog = GetControl32BitValue(control) ? true : false;
			
			controlID.id = iNibXShowTimeInFrz;
			GetControlByID(tWindowRef, &controlID, &control);
			showtimeinfrz = GetControl32BitValue(control) ? true : false;
			
			controlID.id = iNibXMusicBoxMode;
			GetControlByID(tWindowRef, &controlID, &control);
			musicboxmode = GetControl32BitValue(control);

			controlID.id = iNibXEnableToggle;
			GetControlByID(tWindowRef, &controlID, &control);
			enabletoggle = GetControl32BitValue(control) ? true : false;
			
			controlID.id = iNibXAllowEvent;
			GetControlByID(tWindowRef, &controlID, &control);
			allowevent = GetControl32BitValue(control) ? true : false;

			controlID.id = iNibXSaveWindowPos;
			GetControlByID(tWindowRef, &controlID, &control);
			savewindowpos = GetControl32BitValue(control) ? true : false;

			err = RemoveEventHandler(prefHandlerRef);
			DisposeEventHandlerUPP(prefEventProc);
			
			err = RemoveEventHandler(tabHandlerRef);
			DisposeEventHandlerUPP(tabEventProc);
			
			ReleaseWindow(tWindowRef);
		}

		DisposeNibReference(nibRef);
		
		GFX.RealPitch = GFX.Pitch = GFX.Pitch2 = 512 * 2;
		S9xGraphicsInit();
		
		macSoundPitchMaster = macSoundPitch;
		
		switch (mac7110Load)
		{
			case 1:
				LoadUp7110 = &SPC7110Load;
				break;
			
			case 2:
				LoadUp7110 = &SPC7110Grab;
				break;
				
			case 3:
				LoadUp7110 = &SPC7110Open;
		}
		
		cacheMegs = mac7110Megs;

		if (((old16BitPlayback != Settings.SixteenBitSound  ) ||
			 (oldStereo        != Settings.Stereo           ) ||
		     (oldPlaybackRate  != Settings.SoundPlaybackRate) ||
			 (oldMixInterval   != macSoundInterval          ) ||
			 (oldVolume        != macSoundVolume            )) && cartOpen)
			SNES9X_InitSound();
		
		if ((oldEnableSound != Settings.APUEnabled) && cartOpen)
			AppearanceAlert(kAlertCautionAlert, kSoundWarning, kSoundHint);
			
		if ((oldFullscreen != fullscreen) && cartOpen)
		{
			if (fullscreen)
				DeinitGameWindow();
			else
				InitGameWindow();
		}
	}
}

static void SelectItemOfTabControl(ControlRef tabControl, SInt16 index)
{
	ControlRef	userPane, selectedPane = nil;
	ControlID	controlID;
	UInt16		i;
	
	lastTabIndex = index;
	controlID.signature = 'tabs';
	
	for (i = 1; i < tabList[0] + 1; i++)
	{
		controlID.id = tabList[i];
		GetControlByID(GetControlOwner(tabControl), &controlID, &userPane);
		
		if (i == index)
			selectedPane = userPane;
		else
		{
			SetControlVisibility(userPane, false, false);
			DisableControl(userPane);
		}
	}
	
	if (selectedPane != nil)
	{
		EnableControl(selectedPane);
		SetControlVisibility(selectedPane, true, true);
	}
	
	Draw1Control(tabControl);
}

static pascal OSStatus TabEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
	#pragma unused (inHandlerRef, inUserData)
	
	OSStatus	result = eventNotHandledErr;
	ControlRef	theControl;
	ControlID	controlID;
	
	GetEventParameter(inEvent, kEventParamDirectObject, typeControlRef, nil, sizeof(ControlRef), nil, &theControl);
	GetControlID(theControl, &controlID);
	
	if ((controlID.id == 128) && (GetControl32BitValue(theControl) != lastTabIndex))
	{
		SelectItemOfTabControl(theControl, GetControl32BitValue(theControl));
		result = noErr;
	}    
	
	return result;
}

static pascal OSStatus PreferencesEventHandler(EventHandlerCallRef inHandlerRef, EventRef inEvent, void *inUserData)
{
	#pragma unused (inHandlerRef)
	
	OSStatus	err, result = eventNotHandledErr;
	WindowRef	tWindowRef = (WindowRef) inUserData;

	switch (GetEventClass(inEvent))
	{
		case kEventClassWindow:
			switch (GetEventKind(inEvent))
			{
				case kEventWindowClose:
					QuitAppModalLoopForWindow(tWindowRef);
					result = noErr;
			}
			
			break;

		case kEventClassCommand:
			switch (GetEventKind(inEvent))
			{
				case kEventCommandProcess:
					HICommand	tHICommand;

					err = GetEventParameter(inEvent, kEventParamDirectObject, typeHICommand, nil, sizeof(HICommand), nil, &tHICommand);
					if (err == noErr)
					{
						switch (tHICommand.commandID)
						{
							case 'S_EF':
								HideWindow(tWindowRef);
								ConfigureSoundEffects();
								ShowWindow(tWindowRef);
								result = noErr;
						}
					}
			}
	}
	
	return result;
}

void ConfigureHiddenOptions(void)
{
	return;
}
